@@ -17,7 +17,7 @@ def mockKrakenXDevice():
 
 @pytest.fixture
 def mockOldKrakenXDevice():
-    device = _MockKrakenDevice(fw_version=(2, 5, 8))
+    device = _MockKrakenDevice(fw_version=(2, 1, 8))
     dev = Kraken2(device, 'Mock X62', device_type=Kraken2.DEVICE_KRAKENX)
 
     dev.connect()
@@ -72,7 +72,13 @@ def test_kraken_connect(mockKrakenXDevice):
 def test_kraken_initialize(mockKrakenXDevice):
     (fw_ver,) = mockKrakenXDevice.initialize()
 
-    assert fw_ver[1] == '6.0.2'
+    assert fw_ver[1] == '6.2'
+
+
+def test_old_kraken_initialize(mockOldKrakenXDevice):
+    (fw_ver,) = mockOldKrakenXDevice.initialize()
+
+    assert fw_ver[1] == '2.1.8'
 
 
 @pytest.mark.parametrize('has_hwmon,direct_access', [(False, False), (True, True)])
@@ -86,7 +92,6 @@ def test_kraken_get_status_directly(mockKrakenXDevice, has_hwmon, direct_access)
         ('Liquid temperature', pytest.approx(30.9), '°C'),
         ('Fan speed', 1499, 'rpm'),
         ('Pump speed', 2702, 'rpm'),
-        ('Firmware version', '6.0.2', ''),
     ]
 
     assert sorted(got) == sorted(expected)
@@ -104,7 +109,6 @@ def test_kraken_get_status_from_hwmon(mockKrakenXDevice, tmp_path):
         ('Liquid temperature', pytest.approx(20.9), '°C'),
         ('Fan speed', 2499, 'rpm'),
         ('Pump speed', 1702, 'rpm'),
-        ('Firmware version', '6.0.2', ''),
     ]
 
     assert sorted(got) == sorted(expected)
@@ -146,7 +150,7 @@ def test_kraken_speed_profiles_not_supported(mockOldKrakenXDevice):
 
 def test_krakenM_initialize(mockKrakenMDevice):
     (fw_ver,) = mockKrakenMDevice.initialize()
-    assert fw_ver[1] == '6.0.2'
+    assert fw_ver[1] == '6.2'
 
 
 def test_krakenM_get_status(mockKrakenMDevice):
